from .AggregateMatrix import AggregateMatrix as aggregate_matrix
from .Clustering import Clustering as cluster
from .DeAnalysis import DeAnalysis as de_analysis
from .Enrichment import Enrichment as enrichment
from .AnnotateCluster import AnnotateCluster as annotate_cluster
from .Plotting import Plotting as plot
# from .SubClustering import SubClustering as subcluster
from .CheckSampleIndexes import CheckSampleIndexes as check_indexes
from .FindMarkers import FindMarkers as find_markers
from .DownSample import DownSample as down_sample
from .ConvertH5SC import ConvertH5SC as convert_h5sc
